﻿using System;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Controls.Primitives;
using System.Windows.Input;
using Telerik.Windows.Controls;

namespace Telerik.Windows.Media.Imaging.Tools.UI
{
    [TemplatePart(Name = FixedRatioCropAdorner.MainGridName, Type = typeof(Grid))]
    [TemplatePart(Name = FixedRatioCropAdorner.CropSizeLabelName, Type = typeof(TextBlock))]
    [TemplatePart(Name = FixedRatioCropAdorner.ThumbTLName, Type = typeof(Thumb))]
    [TemplatePart(Name = FixedRatioCropAdorner.ThumbTRName, Type = typeof(Thumb))]
    [TemplatePart(Name = FixedRatioCropAdorner.ThumbBRName, Type = typeof(Thumb))]
    [TemplatePart(Name = FixedRatioCropAdorner.ThumbBLName, Type = typeof(Thumb))]
    [TemplatePart(Name = FixedRatioCropAdorner.ThumbLName, Type = typeof(Thumb))]
    [TemplatePart(Name = FixedRatioCropAdorner.ThumbTName, Type = typeof(Thumb))]
    [TemplatePart(Name = FixedRatioCropAdorner.ThumbRName, Type = typeof(Thumb))]
    [TemplatePart(Name = FixedRatioCropAdorner.ThumbBName, Type = typeof(Thumb))]
    [TemplatePart(Name = FixedRatioCropAdorner.ThumbCName, Type = typeof(Thumb))]
    public class FixedRatioCropAdorner : Control
    {
        #region Constants

        public const string MainGridName = "MainGrid";
        public const string CropSizeLabelName = "CropSizeLabel";

        public const string ThumbTLName = "ThumbTL";
        public const string ThumbTRName = "ThumbTR";
        public const string ThumbBRName = "ThumbBR";
        public const string ThumbBLName = "ThumbBL";

        public const string ThumbLName = "ThumbL";
        public const string ThumbTName = "ThumbT";
        public const string ThumbRName = "ThumbR";
        public const string ThumbBName = "ThumbB";
        public const string ThumbCName = "ThumbC";

        #endregion

        private Rect cropRect;
        private Rect startDragCropRect;
        private Point accumulatedOffset;

        private Grid mainGrid;
        private TextBlock cropSizeLabel;
        private Thumb thumbTL, thumbTR, thumbBR, thumbBL;
        private Thumb thumbL, thumbT, thumbR, thumbB, thumbC;

        public event EventHandler CropRectChanged;

        public RadImageEditor ImageEditor { get; set; }

        public Rect CropRect
        {
            get
            {
                return this.cropRect;
            }
            set
            {
                this.cropRect = value;
                this.UpdateGrid(false);
            }
        }

#if WPF
        static FixedRatioCropAdorner()
        {
            DefaultStyleKeyProperty.OverrideMetadata(typeof(FixedRatioCropAdorner), new FrameworkPropertyMetadata(typeof(FixedRatioCropAdorner)));
        }

        internal virtual void SetDefaultStyleKey()
        {
            Theme theme = StyleManager.GetTheme(this);
            this.DefaultStyleKey = ThemeResourceKey.GetDefaultStyleKey(theme, typeof(FixedRatioCropAdorner));
        }

        /// <summary>
        /// Raises the <see cref="E:System.Windows.FrameworkElement.Initialized" /> event.
        /// This method is invoked whenever <see cref="P:System.Windows.FrameworkElement.IsInitialized" />
        /// is set to true internally.
        /// </summary>
        /// <param name="e">The <see cref="T:System.Windows.RoutedEventArgs" /> that contains the event data.</param>
        protected override void OnInitialized(EventArgs e)
        {
            base.OnInitialized(e);
            StyleManager.SetDefaultStyleKey(this, typeof(FixedRatioCropAdorner));
        }
#endif

        public FixedRatioCropAdorner()
        {
#if SILVERLIGHT
            this.DefaultStyleKey = typeof(FixedRatioCropAdorner);
#endif
        }

        public override void OnApplyTemplate()
        {
            base.OnApplyTemplate();

            this.mainGrid = base.GetTemplateChild(FixedRatioCropAdorner.MainGridName) as Grid;
            this.cropSizeLabel = base.GetTemplateChild(FixedRatioCropAdorner.CropSizeLabelName) as TextBlock;

            this.thumbTL = base.GetTemplateChild(FixedRatioCropAdorner.ThumbTLName) as Thumb;
            this.thumbTR = base.GetTemplateChild(FixedRatioCropAdorner.ThumbTRName) as Thumb;
            this.thumbBR = base.GetTemplateChild(FixedRatioCropAdorner.ThumbBRName) as Thumb;
            this.thumbBL = base.GetTemplateChild(FixedRatioCropAdorner.ThumbBLName) as Thumb;

            this.thumbL = base.GetTemplateChild(FixedRatioCropAdorner.ThumbLName) as Thumb;
            this.thumbT = base.GetTemplateChild(FixedRatioCropAdorner.ThumbTName) as Thumb;
            this.thumbR = base.GetTemplateChild(FixedRatioCropAdorner.ThumbRName) as Thumb;
            this.thumbB = base.GetTemplateChild(FixedRatioCropAdorner.ThumbBName) as Thumb;
            this.thumbC = base.GetTemplateChild(FixedRatioCropAdorner.ThumbCName) as Thumb;

            this.thumbTL.DragStarted += this.thumb_DragStarted;
            this.thumbTL.DragDelta += this.thumbTL_DragDelta;
            this.thumbTR.DragStarted += this.thumb_DragStarted;
            this.thumbTR.DragDelta += this.thumbTR_DragDelta;
            this.thumbBR.DragStarted += this.thumb_DragStarted;
            this.thumbBR.DragDelta += this.thumbBR_DragDelta;
            this.thumbBL.DragStarted += this.thumb_DragStarted;
            this.thumbBL.DragDelta += this.thumbBL_DragDelta;

            this.thumbL.DragStarted += this.thumb_DragStarted;
            this.thumbL.DragDelta += this.thumbL_DragDelta;
            this.thumbT.DragStarted += this.thumb_DragStarted;
            this.thumbT.DragDelta += this.thumbT_DragDelta;
            this.thumbR.DragStarted += this.thumb_DragStarted;
            this.thumbR.DragDelta += this.thumbR_DragDelta;
            this.thumbB.DragStarted += this.thumb_DragStarted;
            this.thumbB.DragDelta += this.thumbB_DragDelta;
            this.thumbC.DragStarted += this.thumb_DragStarted;
            this.thumbC.DragDelta += this.thumbC_DragDelta;

            this.UpdateGrid(false);
        }

        private void thumbL_DragDelta(object sender, DragDeltaEventArgs e)
        {
            this.Accumulate(e);

            this.ChangeLeft();

            this.UpdateGrid();
        }

        private void thumbT_DragDelta(object sender, DragDeltaEventArgs e)
        {
            this.Accumulate(e);

            this.ChangeTop();

            this.UpdateGrid();
        }

        private void thumbR_DragDelta(object sender, DragDeltaEventArgs e)
        {
            this.Accumulate(e);

            this.ChangeWidth();

            this.UpdateGrid();
        }

        private void thumbB_DragDelta(object sender, DragDeltaEventArgs e)
        {
            this.Accumulate(e);

            this.ChangeHeight();

            this.UpdateGrid();
        }

        private void thumbC_DragDelta(object sender, DragDeltaEventArgs e)
        {
            this.Accumulate(e);

            this.ChangeC();

            this.UpdateGrid();
        }

        private void thumbTL_DragDelta(object sender, DragDeltaEventArgs e)
        {
            this.Accumulate(e);

            this.ChangeLeft();
            this.ChangeTop();

            this.UpdateGrid();
        }

        private void thumbTR_DragDelta(object sender, DragDeltaEventArgs e)
        {
            this.Accumulate(e);

            this.ChangeWidth();
            this.ChangeTop();

            this.UpdateGrid();
        }

        private void thumbBR_DragDelta(object sender, DragDeltaEventArgs e)
        {
            this.Accumulate(e);

            this.ChangeWidth();
            this.ChangeHeight();

            this.UpdateGrid();
        }

        private void thumbBL_DragDelta(object sender, DragDeltaEventArgs e)
        {
            this.Accumulate(e);

            this.ChangeLeft();
            this.ChangeHeight();

            this.UpdateGrid();
        }

        private void ChangeWidth()
        {
            this.cropRect.Width = Limit(
                1,
                this.ImageEditor.Image.Width - this.cropRect.X,
                this.startDragCropRect.Width + this.accumulatedOffset.X);
        }

        private void ChangeHeight()
        {
            this.cropRect.Height = Limit(
                1,
                this.ImageEditor.Image.Height - this.cropRect.Y,
                this.startDragCropRect.Height + this.accumulatedOffset.Y);
        }

        private void ChangeLeft()
        {
            double change = Limit(
                -startDragCropRect.X,
                startDragCropRect.Width - 1,
                this.accumulatedOffset.X
            );

            this.cropRect.X = startDragCropRect.X + change;
            this.cropRect.Width = startDragCropRect.Width - change;
        }

        private void ChangeTop()
        {
            double change = Limit(
                -startDragCropRect.Y,
                startDragCropRect.Height - 1,
                this.accumulatedOffset.Y
            );

            this.cropRect.Y = startDragCropRect.Y + change;
            this.cropRect.Height = startDragCropRect.Height - change;
        }

        private void ChangeC()
        {
            double changeX = Limit(
                -startDragCropRect.X,
                this.ImageEditor.Image.Width - this.startDragCropRect.Width - this.startDragCropRect.X,
                this.accumulatedOffset.X
            );
            double changeY = Limit(
                -startDragCropRect.Y,
                this.ImageEditor.Image.Height - this.startDragCropRect.Height - this.startDragCropRect.Y,
                this.accumulatedOffset.Y
            );

            this.cropRect.X = startDragCropRect.X + changeX;
            this.cropRect.Y = startDragCropRect.Y + changeY;            
        }

        protected virtual void OnCropRectChanged()
        {
            if (this.CropRectChanged != null)
            {
                this.CropRectChanged(this, EventArgs.Empty);
            }
        }

#if SILVERLIGHT
        private void StartAccumulate() {
        }
#else
        private Point lastMousePosition;

        private void StartAccumulate()
        {
            this.lastMousePosition = Mouse.GetPosition(this);
        }
#endif

#if SILVERLIGHT
        private void Accumulate(DragDeltaEventArgs e)
        {
            this.accumulatedOffset.X = this.accumulatedOffset.X + e.HorizontalChange / this.ImageEditor.ActualScaleFactor;
            this.accumulatedOffset.Y = this.accumulatedOffset.Y + e.VerticalChange / this.ImageEditor.ActualScaleFactor;
        }
#else
        private void Accumulate(DragDeltaEventArgs e)
        {
            Point mousePosition = Mouse.GetPosition(this);
            this.accumulatedOffset.X = this.accumulatedOffset.X + (mousePosition.X - this.lastMousePosition.X) / this.ImageEditor.ActualScaleFactor;
            this.accumulatedOffset.Y = this.accumulatedOffset.Y + (mousePosition.Y - this.lastMousePosition.Y) / this.ImageEditor.ActualScaleFactor;
            this.lastMousePosition = mousePosition;
        }
#endif

        private void thumb_DragStarted(object sender, DragStartedEventArgs e)
        {
            this.StartAccumulate();

            this.startDragCropRect = this.cropRect;
            this.accumulatedOffset = new Point(0, 0);
        }

        private static double Limit(double min, double max, double value)
        {
            if (value < min) return min;
            if (value > max) return max;
            return value;
        }

        private void UpdateGrid(bool onCropRectChanged = true)
        {
            if (this.mainGrid == null)
            {
                return;
            }

            double x = this.cropRect.X;
            double y = this.cropRect.Y;
            double w = this.cropRect.Width;
            double h = this.cropRect.Height;

            double lh = Math.Max(0, this.ImageEditor.Image.Height - h - y);
            double lw = Math.Max(0, this.ImageEditor.Image.Width - w - x);

            this.mainGrid.RowDefinitions[0].Height = new GridLength(y, GridUnitType.Star);
            this.mainGrid.RowDefinitions[1].Height = new GridLength(h, GridUnitType.Star);
            this.mainGrid.RowDefinitions[2].Height = new GridLength(lh, GridUnitType.Star);
            this.mainGrid.ColumnDefinitions[0].Width = new GridLength(x, GridUnitType.Star);
            this.mainGrid.ColumnDefinitions[1].Width = new GridLength(w, GridUnitType.Star);
            this.mainGrid.ColumnDefinitions[2].Width = new GridLength(lw, GridUnitType.Star);

            this.cropSizeLabel.Text = Math.Round(w) + " x " + Math.Round(h);

            if (onCropRectChanged)
            {
                this.OnCropRectChanged();
            }
        }
    }
}
